import os
import os.path as op
import click
import shutil

from ..context import pass_context, on_finished

NXC_FOLDER_PATH = op.abspath(op.join(op.dirname(__file__), "../../nxc"))
NXC_JSON_PATH = op.abspath(op.join(op.dirname(__file__), "../../nxc.json"))


@click.command("clean")
@pass_context
@on_finished(lambda ctx: ctx.state.dump())
def cli(ctx):
    """Clean the nxc folder and nxc.json file"""
    if os.path.isfile(NXC_JSON_PATH):
        os.remove(NXC_JSON_PATH)

    if os.path.isdir(NXC_FOLDER_PATH):
        shutil.rmtree(NXC_FOLDER_PATH)



import os
import os.path as op
import json
import subprocess
import click
import copy

from ..flavour import Flavour
from ..actions import read_compose_info
from ..driver.logger import rootlog
from ..driver.machine import Machine
from ..default_role import DefaultRole

from typing import Tuple, Optional


def generate_docker_compose_file(ctx):
    base_docker_compose = ctx.compose_info["docker-compose-file"]
    docker_compose_content = {"services": {}}
    nodes_info = {}

    with open(base_docker_compose) as dc_file:
        dc_json = json.load(dc_file)
        if ctx.roles_quantities == {}:
            roles_quantities = {role: 1 for role in ctx.compose_info["nodes"]}
        else:
            roles_quantities = dict(
                filter(
                    lambda x: x[0] in ctx.compose_info["nodes"],
                    ctx.roles_quantities.items(),
                )
            )
        # Add bind  deployment file inside containers
        deployment_file = op.join(
            ctx.envdir, f"deploy/{ctx.composition_flavour_prefix}.json"
        )
        for service in dc_json["services"].values():
            (service["volumes"]).append(
                {
                    "type": "bind",
                    "source": deployment_file,
                    "target": "/etc/nxc/deployment.json",
                }
            )
        for role, quantities in roles_quantities.items():
            if type(quantities) is int:
                if quantities == 1:
                    hostname = f"{role}"
                    config = copy.copy(dc_json["services"][role])
                    config["hostname"] = hostname
                    docker_compose_content["services"][hostname] = config
                    nodes_info[hostname] = role
                else:
                    for i in range(1, quantities + 1):
                        hostname = f"{role}{i}"
                        config = copy.copy(dc_json["services"][role])
                        config["hostname"] = hostname
                        docker_compose_content["services"][hostname] = config
                        nodes_info[hostname] = role
            elif type(quantities) is list:
                for hostname in quantities:
                    config = copy.copy(dc_json["services"][role])
                    config["hostname"] = hostname
                    docker_compose_content["services"][hostname] = config
                    nodes_info[hostname] = role
            elif type(quantities) is DefaultRole:
                nb_min_nodes = quantities.nb_min_nodes
                ctx.log(f"Docker: Using DefaultRole -> {nb_min_nodes} nodes for role '{role}'")
                if nb_min_nodes == 1:
                    hostname = f"{role}"
                    config = copy.copy(dc_json["services"][role])
                    config["hostname"] = hostname
                    docker_compose_content["services"][hostname] = config
                    nodes_info[hostname] = role
                else:
                    for i in range(1, nb_min_nodes + 1):
                        hostname = f"{role}{i}"
                        config = copy.copy(dc_json["services"][role])
                        config["hostname"] = hostname
                        docker_compose_content["services"][hostname] = config
                        nodes_info[hostname] = role
            else:
                raise Exception("Unvalid type for specifying the roles of the nodes")
        docker_compose_content["version"] = dc_json["version"]
        docker_compose_content["x-nxc"] = dc_json["x-nxc"]


    deploy_dir = op.join(ctx.envdir, "deploy", "docker_compose")
    if not op.exists(deploy_dir):
        create = click.style("   create", fg="green")
        ctx.log("   " + create + "  " + deploy_dir)
        os.mkdir(deploy_dir)
    docker_compose_path = op.join(deploy_dir, "docker-compose.json")

    with open(docker_compose_path, "w") as outfile:
        outfile.write(json.dumps(docker_compose_content))
    return docker_compose_path, nodes_info


def generate_deployment_info_docker(ctx):
    if not ctx.compose_info:
        read_compose_info(ctx)

    deploy_dir = op.join(ctx.envdir, "deploy")
    if not op.exists(deploy_dir):
        create = click.style("   create", fg="green")
        ctx.log("   " + create + "  " + deploy_dir)
        os.mkdir(deploy_dir)

    docker_compose_path, nodes_info = generate_docker_compose_file(ctx)
    deployment = {
        # "nodes": ctx.compose_info["nodes"],
        "nodes": list(nodes_info.keys()),
        # "deployment": {n: {"role": n} for n in ctx.compose_info["nodes"]},
        "deployment": {node_name: {"role": role_name} for (node_name, role_name) in nodes_info.items()},
        "docker-compose-file": docker_compose_path,
    }

    if "all" in ctx.compose_info:
        deployment["all"] = ctx.compose_info["all"]

    json_deployment = json.dumps(deployment, indent=2)

    with open(
        op.join(deploy_dir, f"{ctx.composition_flavour_prefix}.json"), "w"
    ) as outfile:
        outfile.write(json_deployment)

    ctx.deployment_info = deployment
    return docker_compose_path


class DockerFlavour(Flavour):

    docker_compose_file = None

    def __init__(self, ctx):
        super().__init__(ctx)

        self.name = "docker"
        self.description = ""
        # TOR self.docker_processes = {}

    def generate_deployment_info(self):
        self.docker_compose_file = generate_deployment_info_docker(self.ctx)

    def driver_initialize(self, tmp_dir):

        assert self.ctx.deployment_info
        if not self.docker_compose_file:
            self.docker_compose_file = self.ctx.deployment_info["docker-compose-file"]

        nodes_names = self.ctx.deployment_info["nodes"]
        for name in nodes_names:
            self.machines.append(
                Machine(self.ctx, tmp_dir=tmp_dir, start_command="", name=name,)
            )

    def check(self, state="running"):
        check_process = subprocess.check_output(
            [
                "docker-compose",
                "-f",
                self.docker_compose_file,
                "ps",
                "--services",
                "--filter",
                f"status={state}",
            ],
        )
        return len(check_process.decode().rstrip("\n").splitlines())

    def connect(self, machine):
        if machine.connected:
            return
        self.start_all()

    def start_all(self):
        if not self.external_connect:
            with rootlog.nested("starting docker-compose"):
                subprocess.Popen(
                    ["docker-compose", "-f", self.docker_compose_file, "up", "-d"]
                )

            self.wait_on_check()

        for machine in self.machines:
            if not machine.connected:
                self.start(machine)
                machine.connected = True

    def start(self, machine):  # TODO MOVE to Connect ???
        assert machine.name
        assert self.docker_compose_file

        machine.start_process_shell(
            [
                "docker-compose",
                "-f",
                self.docker_compose_file,
                "exec",
                "-T",
                machine.name,
                "bash",
            ]
        )

    def execute(
        self,
        machine,
        command: str,
        check_return: bool = True,
        timeout: Optional[int] = 900,
    ) -> Tuple[int, str]:

        return machine.execute_process_shell(command, check_return, timeout)

    def restart(self, machine):
        machine.restart_process_shell()

    def cleanup(self):
        # TODO handle stdout/stderr
        if not self.docker_compose_file:
            self.docker_compose_file = self.ctx.deployment_info["docker-compose-file"]
        subprocess.Popen(
            [
                "docker-compose",
                "-f",
                self.docker_compose_file,
                "down",
                "--remove-orphans",
            ]
        )

    def shell_interact(self, machine) -> None:
        self.connect(machine)
        self.ext_connect("root", machine.name)

    def ext_connect(self, user, node, execute=True):
        if not self.docker_compose_file:
            self.docker_compose_file = self.ctx.deployment_info["docker-compose-file"]

        cmd = f"docker-compose -f {self.docker_compose_file} exec -u {user} {node} bash"

        if execute:
            return_code = subprocess.run(cmd, shell=True).returncode

            if return_code:
                self.ctx.wlog(f"Docker exit code is not null: {return_code}")
            return return_code
        else:
            return cmd



from ..flavour import Flavour
from ..actions import generate_deployment_info, ssh_connect


class NixosTestFlavour(Flavour):
    def __init__(self, ctx):
        super().__init__(ctx)
        self.name = "nixos-test"


class NixosTestDriverFlavour(Flavour):
    def __init__(self, ctx):
        super().__init__(ctx)
        self.name = "nixos-test-driver"


class NixosTestSshFlavour(Flavour):
    def __init__(self, ctx):
        super().__init__(ctx)
        self.name = "nixos-test-ssh"

    def generate_deployment_info(self):
        generate_deployment_info(self.ctx)

    def ext_connect(self, user, node, execute=True):
        return ssh_connect(self.ctx, user, node, execute)



import os
import sys
import json
import base64
import subprocess
from ..httpd import HTTPDaemon

from ..flavour import Flavour
from ..actions import generate_deployment_info, ssh_connect, kill_proc_tree
from ..driver.vlan import VLan
from ..driver.logger import rootlog
from ..driver.machine import Machine, StartScript


class VmRamdiskFlavour(Flavour):
    """
    The Vm Ramdisk flavour. This is flavour provides a system image to be executed with QEMU and use memory only for root system. By consequence lot of ram is used around 2Go minimum by node.
    """

    vm = True
    tmp_dir = None
    vlan = None

    def __init__(self, ctx):
        super().__init__(ctx)

        self.name = "vm-ramdisk"
        self.image = {"type": "ramdisk", "distribution": "all-in-one"}
        self.description = "Plain vm ramdisk (all-in-memory), need lot of ram !"

    def generate_deployment_info(self):
        generate_deployment_info(self.ctx)

    def driver_initialize(self, tmp_dir):
        self.tmp_dir = tmp_dir

        ctx = self.ctx
        deployment = ctx.deployment_info

        if ctx.no_start:  #
            deployment_nodes = self.ctx.deployment_info["deployment"]
            for ip, node in deployment_nodes.items():
                self.machines.append(
                    Machine(
                        self.ctx,
                        ip=ip,
                        tmp_dir=tmp_dir,
                        start_command="",
                        keep_vm_state=False,
                        name=node["host"],
                    )
                )

            for machine in self.machines:
                if not machine.connected:
                    self.start(machine)
                machine.connected = True
            return

        os.environ["KERNEL"] = deployment["all"]["kernel"]
        os.environ["INITRD"] = deployment["all"]["initrd"]
        base_qemu_script = deployment["all"]["qemu_script"]

        debug_stage1 = None
        debug_var_base = ""
        if "DEBUG_STAGE1" in os.environ:
            debug_stage1 = os.environ["DEBUG_STAGE1"]

        if "DEPLOY" not in os.environ:
            if ctx.use_httpd:
                if not ctx.httpd:
                    ctx.httpd = HTTPDaemon(ctx=ctx)
                    ctx.httpd.start(directory=ctx.envdir)
                base_url = f"http://10.0.2.2:{ctx.httpd.port}"
                deploy_info_src = (
                    f"{base_url}/deploy/{ctx.composition_flavour_prefix}.json"
                )
            else:
                if not ctx.deployment_info_b64:
                    deployment_info_str = json.dumps(deployment)
                    deploy_info_src = base64.b64encode(
                        deployment_info_str.encode()
                    ).decode()
                else:
                    deploy_info_src = ctx.deployment_info_b64
                if len(deploy_info_src) > (4096 - 256):
                    rootlog.nested(
                        "The base64 encoded deploy data is too large: use an http server to serve it"
                    )
                    sys.exit(1)
            os.environ["DEPLOY"] = f"deploy={deploy_info_src}"
            print(f"deploy={deploy_info_src}")
        else:
            rootlog.nested(f'Variable environment DEPLOY: {os.environ["DEPLOY"]}')

        for var_env in ["QEMU_APPEND", "DEBUG_INITRD"]:
            if var_env in os.environ:
                rootlog.nested(f"Variable environment {var_env}: {os.environ[var_env]}")

        if debug_stage1:
            debug_var_base = f'KERNEL={deployment["all"]["kernel"]} \\\nINITRD={deployment["all"]["initrd"]} \\\n'
        base_qemu_script = deployment["all"]["qemu_script"]

        ip_addresses = []
        for i in range(len(deployment["deployment"])):
            ip = "10.0.2.{}".format(15 + i)
            ip_addresses.append(ip)
            if ip not in deployment["deployment"]:
                rootlog.nested(f"In vm mode, {ip} must be present")
                sys.exit(1)
            v = deployment["deployment"][ip]

            name = v["host"]

            if base_qemu_script:
                qemu_script = base_qemu_script
            else:
                qemu_script = v["qemu_script"]

            if debug_stage1:
                if debug_stage1 == name:
                    # debloy="DEPLOY=deploy=http://10.0.2.1:8000/deploy/composition::vm-ramdisk.json \\\n"
                    # deploy="DEPLOY=deploy={ctx.deploy_info_src}"
                    params = f'{debug_var_base}INIT={v["init"]} \\\n'
                    debug = " DEBUG_INITRD=boot.debug1mounts "
                    # params = f'{params}QEMU_VDE_SOCKET={self.vlan.socket_dir}{debug}VM_ID={v["vm_id"]} ROLE={name}\\\n'
                    params = f'{params}QEMU_VDE_SOCKET={self.vlan.socket_dir}{debug}VM_ID={v["vm_id"]} \\\n'
                    if "DEPLOY" in os.environ:
                        params = f'DEPLOY={os.environ["DEPLOY"]} \\\n{params}'
                    print()
                    print(f"DEBUG STAGE1 on role: {name}")
                    print(f"{params}{qemu_script}")
                    # subprocess.call(f"{params} bash -x {qemu_script}", shell=True)
                    subprocess.call(f"{params}{qemu_script}", shell=True)
                    sys.exit(0)
            else:

                self.machines.append(
                    Machine(
                        self.ctx,
                        tmp_dir=tmp_dir,
                        start_command=StartScript(qemu_script, v["vm_id"], self),
                        name=name,
                        ip=ip,
                        keep_vm_state=False,
                        vm_id=v["vm_id"],
                        init=v["init"],
                    )
                )

    def vlan(self):
        self.vlan = VLan(0, self.tmp_dir, tap0=True)
        return self.vlan

    def start(self, machine):
        if not self.ctx.no_start:
            machine._start_vm()
        else:
            machine.start_process_shell(
                [
                    "ssh",
                    "-t",
                    "-o",
                    "StrictHostKeyChecking=no",
                    "-l",
                    "root",
                    machine.ip,
                ]
            )

    def release(self, machine):
        if machine.pid is None:
            return
        rootlog.info(f"kill machine (pid {machine.pid})")
        assert machine.process
        assert machine.shell
        assert machine.monitor
        assert machine.serial_thread

        # Kill children
        kill_proc_tree(machine.pid, include_parent=False)

        machine.process.terminate()
        machine.shell.close()
        machine.monitor.close()
        machine.serial_thread.join()

    def ext_connect(self, user, node, execute=True):
        return ssh_connect(self.ctx, user, node, execute)

    # def start_all(self):
    #    with rootlog.nested("start all VMs"):

    # def start_composition(self):
    #     pass

    # def driver_start(self):
    #     pass



from contextlib import contextmanager
from pathlib import Path
from typing import Any, Dict, Iterator, List
import os
import tempfile
import signal

from .logger import rootlog
from .machine import Machine, retry
from .vlan import VLan
from ..flavours import use_flavour_method_if_any


class Driver:
    """A handle to the driver that sets up the environment
    and runs the tests"""

    tests: str
    vlans: List[VLan]
    machines: List[Machine]

    def __init__(
        self,
        ctx,
        start_scripts: List[str],
        vlans: List[int],
        tests: str,
        keep_vm_state: bool = False,
    ):
        self.ctx = ctx
        self.tests = tests

        tmp_dir = Path(os.environ.get("TMPDIR", tempfile.gettempdir()))
        tmp_dir.mkdir(mode=0o700, exist_ok=True)

        # if hasattr(ctx.flavour, "vlan"):
        #    self.vlans = [ctx.flavour.vlan()]

        ctx.flavour.driver_initialize(tmp_dir)

        # if hasattr(ctx.flavour, "machines") and ctx.flavour.machines:
        if ctx.flavour.machines:
            self.machines = ctx.flavour.machines
        else:
            rootlog.warning("No machine defined during driver init")

        if hasattr(ctx.flavour, "vlan") and not ctx.no_start:
            self.vlans = [ctx.flavour.vlan()]

        # def cmd(scripts: List[str]) -> Iterator[NixStartScript]:
        #     for s in scripts:
        #         yield NixStartScript(s)

        # self.machines = [
        #     Machine(
        #         start_command=cmd,
        #         keep_vm_state=keep_vm_state,
        #         name=cmd.machine_name,
        #         tmp_dir=tmp_dir,
        #     )
        #     for cmd in cmd(start_scripts)
        # ]

    def __enter__(self) -> "Driver":
        return self

    def __exit__(self, *_: Any) -> None:
        if not self.ctx.interactive and not self.ctx.execute_test_script:
            if self.ctx.sigwait:
                with rootlog.nested("wait any signal to exit"):
                    signal.sigwait(signal.valid_signals())
                    self.cleanup()
        else:
            self.cleanup()

    @use_flavour_method_if_any
    def cleanup(self):
        with rootlog.nested("cleanup"):
            for machine in self.machines:
                machine.release()

    def subtest(self, name: str) -> Iterator[None]:
        """Group logs under a given test name"""
        with rootlog.nested(name):
            try:
                yield
                return True
            except Exception as e:
                rootlog.error(f'Test "{name}" failed with error: "{e}"')
                raise e

    def test_symbols(self) -> Dict[str, Any]:
        @contextmanager
        def subtest(name: str) -> Iterator[None]:
            return self.subtest(name)

        general_symbols = dict(
            start_all=self.start_all,
            test_script=self.test_script,
            machines=self.machines,
            # vlans=self.vlans,
            driver=self,
            log=rootlog,
            os=os,
            subtest=subtest,
            run_tests=self.run_tests,
            join_all=self.join_all,
            retry=retry,
            serial_stdout_off=self.serial_stdout_off,
            serial_stdout_on=self.serial_stdout_on,
            Machine=Machine,  # for typing
        )
        machine_symbols = {m.name: m for m in self.machines}
        # If there's exactly one machine, make it available under the name
        # "machine", even if it's not called that.
        if len(self.machines) == 1:
            (machine_symbols["machine"],) = self.machines

        # vlan_symbols = {
        #    f"vlan{v.nr}": self.vlans[idx] for idx, v in enumerate(self.vlans)
        # }
        print(
            "additionally exposed symbols:\n    "
            + ", ".join(map(lambda m: m.name, self.machines))
            + ",\n    "
            # + ", ".join(map(lambda v: f"vlan{v.nr}", self.vlans))
            + ",\n    "
            + ", ".join(list(general_symbols.keys()))
        )
        return {**general_symbols, **machine_symbols}  # , **vlan_symbols}

    def test_script(self) -> None:
        """Run the test script"""
        with rootlog.nested("run the test script"):
            symbols = self.test_symbols()  # call eagerly
            exec(self.tests, symbols, None)

    def run_tests(self) -> None:
        """Run the test script (for non-interactive test runs)"""
        self.test_script()
        # TODO: Collect coverage data
        for machine in self.machines:
            if machine.is_up():
                machine.execute("sync")

    @use_flavour_method_if_any
    def start_all(self) -> None:
        """Start all machines"""
        with rootlog.nested("start all VMs"):
            for machine in self.machines:
                machine.start()

    def join_all(self) -> None:
        """Wait for all machines to shut down"""
        with rootlog.nested("wait for all VMs to finish"):
            for machine in self.machines:
                machine.wait_for_shutdown()

    def serial_stdout_on(self) -> None:
        rootlog._print_serial_logs = True

    def serial_stdout_off(self) -> None:
        rootlog._print_serial_logs = False



import json
import os.path as op
import pprint

from io import open
from .utils import touch


# TOREMOVE
class State(dict):
    DEFAULTS = {"built": False, "started": False}

    def __init__(self, ctx, state_file):
        dict.__init__(self, self.DEFAULTS)
        self.ctx = ctx
        self.state_file = state_file
        self.load()

    def load(self):
        if op.isfile(self.state_file):
            try:
                with open(self.state_file, "rt") as json_file:
                    self.update(json.loads(json_file.read()))
            except Exception:
                pass

    def dump(self):
        if op.isdir(self.ctx.envdir):
            touch(self.state_file)
            if op.isdir(op.dirname(self.state_file)):
                with open(self.state_file, "w", encoding="utf8") as json_file:
                    json_file.write(json.dumps(self, ensure_ascii=False))

    def __str__(self):
        return pprint.pprint(self)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nixos_compose',
 'nixos_compose.commands',
 'nixos_compose.driver',
 'nixos_compose.flavours',
 'nixos_compose.tools']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'click>=7.1.2',
 'execo>=2.6.8,<3.0.0',
 'halo>=0.0.31,<0.0.32',
 'pexpect>=4.8.0,<5.0.0',
 'psutil>=5.9.0,<6.0.0',
 'ptpython>=3.0.7,<4.0.0',
 'pyinotify>=0.9.6,<0.10.0',
 'requests>=2.27.1,<3.0.0',
 'tomlkit>=0.10.0,<0.11.0']

entry_points = \
{'console_scripts': ['nixos-compose = nixos_compose.cli:main',
                     'nxc = nixos_compose.cli:main']}

setup_kwargs = {
    'name': 'nixos-compose',
    'version': '0.2.4',
    'description': '',
    'long_description': None,
    'author': 'Olivier Richard',
    'author_email': 'olivier.richard@imag.fr',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import os
import os.path as op
import sys
import subprocess
import click
import json

from ..actions import get_nix_command, realpath_from_store
from ..context import pass_context, on_started, on_finished
from ..platform import platform_detection
from ..setup import apply_setup

# FLAVOURS_PATH = op.abspath(op.join(op.dirname(__file__), "../", "flavours"))
# FLAVOURS = os.listdir(FLAVOURS_PATH)


@click.command("build")
@click.argument(
    "composition_file", required=False, type=click.Path(exists=True, resolve_path=True)
)
@click.option(
    "--nix-flags",
    type=click.STRING,
    help='add nix flags (aka options) to nix build command, --nix-flags "--impure"',
)
@click.option("--out-link", "-o", help="path of the symlink to the build result")
@click.option(
    "-f", "--flavour", type=click.STRING, help="Use particular flavour (name or path)",
)
@click.option(
    "-F", "--list-flavours", is_flag=True, help="List available flavour",
)
# TOREMOVE
# @click.option(
#    "--copy-from-store",
#    "-c",
#    is_flag=True,
#    help="Copy artifacts (initrd, kernels, ...) from Nix store to artifact directory",
# )
@click.option("--show-trace", is_flag=True, help="Show Nix trace")
@click.option(
    "--dry-run", is_flag=True, help="Show what this command would do without doing it"
)
@click.option(
    "--dry-build",
    is_flag=True,
    help="Eval build expression and show store entry without building derivation",
)
@click.option(
    "-C",
    "--composition-flavour",
    type=click.STRING,
    help="Use to specify which composition and flavour combinaison to build when muliple compostions are describe at once (see -L options to list them).",
)
# @click.option(
#    "-c", "--composition", type=click.STRING,
#    help="Use to specify which composition to built when muliple compostions are describe at once."
# )
@click.option(
    "-L",
    "--list-compositions-flavours",
    is_flag=True,
    help="List available combinaisons of compositions and flavours",
)
@click.option(
    "-s", "--setup", type=click.STRING, help="Select setup variant",
)
@click.option(
    "-p",
    "--setup-param",
    type=click.STRING,
    multiple=True,
    help="Override setup parameter",
)
@pass_context
@on_finished(lambda ctx: ctx.show_elapsed_time())
@on_started(lambda ctx: ctx.assert_valid_env())
def cli(
    ctx,
    composition_file,
    nix_flags,
    out_link,
    flavour,
    list_flavours,
    show_trace,
    dry_run,
    dry_build,
    composition_flavour,
    list_compositions_flavours,
    setup,
    setup_param,
):
    """Build multi Nixos composition.
    Typically it performs the kind of following command:
      nix build
    """

    def determine_flavour(ctx):
        if "default_flavour" in ctx.nxc and ctx.nxc["default_flavour"]:
            flavour = ctx.nxc["default_flavour"]
        else:
            platform_detection(ctx)
            if ctx.platform:
                flavour = ctx.platform.default_flavour
            else:
                flavour = "default"
        ctx.vlog(f"Seleced flavour: {flavour}")
        return flavour

    if setup or op.exists(op.join(ctx.envdir, "setup.toml")):
        nix_flags, composition_file, composition_flavour, flavour = apply_setup(
            ctx,
            setup,
            nix_flags,
            composition_file,
            composition_flavour,
            flavour,
            setup_param,
        )

    build_cmd = []

    # Do we are in flake context
    if not op.exists(op.join(ctx.envdir, "flake.nix")):
        ctx.elog("Not Found flake.nix file")
        sys.exit(1)

    nix_cmd_base = get_nix_command(ctx)

    description_flavours = get_flavours(nix_cmd_base, ctx)

    flavours = list(description_flavours.keys())

    if list_flavours:
        ctx.log("Flavours List:")
        for k in flavours:
            click.echo(f"{k: <18}: {description_flavours[k]['description']}")
        sys.exit(0)

    # if flavour:
    #    if flavour not in flavours and not op.isfile(flavour):
    #        ctx.elog(f'"{flavour}" is neither a supported flavour nor flavour_path')
    # w        sys.exit(1)

    if not composition_file:
        composition_file = ctx.nxc["composition"]

    if list_compositions_flavours:
        cmd = nix_cmd_base + ["flake", "show", "--json"]
        raw_compositions_flavours = json.loads(
            subprocess.check_output(cmd, cwd=ctx.envdir).decode()
        )
        for compo_flavour in filter(
            lambda x: x not in ["flavoursJson", "showFlavours"],
            raw_compositions_flavours["packages"]["x86_64-linux"].keys(),
        ):
            print(compo_flavour)
        print(
            click.style("Default", fg="green")
            + ": "
            + raw_compositions_flavours["defaultPackage"]["x86_64-linux"]["name"]
        )
        sys.exit(0)

    if show_trace:
        build_cmd += ["--show-trace"]

    if not out_link:
        build_path = op.join(ctx.envdir, "build")
        if not op.exists(build_path):
            create = click.style("   create", fg="green")
            ctx.log("   " + create + "  " + build_path)
            os.mkdir(build_path)

        if not flavour:
            flavour = determine_flavour(ctx)

        if composition_flavour:
            ctx.composition_flavour_prefix = composition_flavour
            ctx.flavour_name = composition_flavour.split("::")[-1]
        else:
            composition_name = (os.path.basename(composition_file)).split(".")[0]
            ctx.composition_name = composition_name
            ctx.flavour_name = flavour
            ctx.composition_flavour_prefix = f"{composition_name}::{flavour}"

        out_link = op.join(build_path, ctx.composition_flavour_prefix)

    if not flavour:
        flavour = determine_flavour(ctx)

    if dry_build:
        build_cmd = nix_cmd_base + ["eval"] + build_cmd + ["--raw"]
    else:
        build_cmd = nix_cmd_base + ["build"] + build_cmd
        if out_link:
            build_cmd += ["-o", out_link]

    if not composition_flavour and flavour:
        composition_flavour = f"composition::{flavour}"
    if flavour:
        build_cmd += [f".#packages.x86_64-linux.{composition_flavour}"]

    # add additional nix flags if any
    if nix_flags:
        build_cmd += nix_flags.split()

    if not dry_run:
        ctx.glog("Starting Build")
        ctx.vlog(build_cmd)
        returncode = subprocess.call(build_cmd, cwd=ctx.envdir)
        if returncode:
            ctx.elog(f"Build return code: {returncode}")
            sys.exit(returncode)

        # Loading the docker image"
        if flavour == "docker" and not dry_build:
            with open(out_link, "r") as compose_info_json:
                content = json.load(compose_info_json)
                docker_image = content["image"]
                docker_load_command = f"docker load < {docker_image}"
                returncode = subprocess.call(docker_load_command, shell=True)
                if returncode:
                    ctx.elog(f"Build return code: {returncode}")
                    sys.exit(returncode)
            ctx.glog("Docker Image loaded")

        ctx.glog("\nBuild completed")
    else:
        ctx.log("Dry-run:")
        ctx.log(f"   working directory:          {ctx.envdir}")
        ctx.log(f"   composition flavour prefix: {ctx.composition_flavour_prefix}")
        ctx.log(f"   build command:              {' '.join(build_cmd)}")


def get_flavours(nix_cmd_base, ctx):
    """
    Returns the json representation of the available flavours
    """
    FLAVOURS_JSON = op.abspath(
        op.join(op.dirname(__file__), "../../nix", "flavours.json")
    )

    flake_location = "."
    output_json = "/tmp/.flavours.json"

    ctx.log("Build list of flavours")
    retcode = subprocess.call(
        nix_cmd_base + ["build", f"{flake_location}#flavoursJson", "-o", output_json],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    if retcode:
        output_json = FLAVOURS_JSON
    else:
        if not op.exists(output_json):
            output_json = realpath_from_store(ctx, output_json)

    return json.load(open(output_json, "r"))



from pyspark.sql import Row, SparkSession
from pyspark.sql import functions as F
from pyspark.sql.functions import udf
from pyspark.sql.types import *
from pyspark.sql.functions import explode


def explode_col(weight):
    return int(weight // 10) * [10.0] + ([] if weight % 10 == 0 else [weight % 10])


spark = SparkSession.builder.getOrCreate()

dataSchema = [
    StructField("feature_1", FloatType()),
    StructField("feature_2", FloatType()),
    StructField("bias_weight", FloatType()),
]

data = [Row(0.1, 0.2, 10.32), Row(0.32, 1.43, 12.8), Row(1.28, 1.12, 0.23)]

df = spark.createDataFrame(spark.sparkContext.parallelize(data), StructType(dataSchema))

normalizing_constant = 100
sum_bias_weight = df.select(F.sum("bias_weight")).collect()[0][0]
normalizing_factor = normalizing_constant / sum_bias_weight
df = df.withColumn("normalized_bias_weight", df.bias_weight * normalizing_factor)
df = df.drop("bias_weight")
df = df.withColumnRenamed("normalized_bias_weight", "bias_weight")

my_udf = udf(lambda x: explode_col(x), ArrayType(FloatType()))
df1 = df.withColumn("explode_val", my_udf(df.bias_weight))
df1 = df1.withColumn("explode_val_1", explode(df1.explode_val)).drop("explode_val")
df1 = df1.drop("bias_weight").withColumnRenamed("explode_val_1", "bias_weight")

df1.show()

assert df1.count() == 12



from nixos_compose.nxc_execo import get_oar_job_nodes_nxc, build_nxc_execo

from execo import Process, SshProcess, Remote
from execo_g5k import oardel, oarsub, OarSubmission, get_oar_job_nodes
from execo_engine import Engine, logger, ParamSweeper, sweep

import sys
import os


class MyEngine(Engine):
    def __init__(self):
        super(MyEngine, self).__init__()
        parser = self.args_parser
        parser.add_argument('--nxc_build_file', help='Path to the NXC deploy file')
        parser.add_argument('--build', action='store_true', help='Either to build the composition')
        parser.add_argument('--nxc_folder', default=f"{os.getcwd()}", help='Path to the nxc folder')
        self.nodes = {}
        self.oar_job_id = -1

    def init(self):
        nb_nodes = 2
        site = "grenoble"
        cluster = "dahu"

        nxc_build_file = None
        if self.args.build:
            (nxc_build_file, _time, _size) = build_nxc_execo(self.args.nxc_folder, site, cluster, walltime=15*60, extra_job_type=["day"])
        elif self.args.nxc_build_file is not None:
            nxc_build_file = self.args.nxc_build_file
        else:
            raise Exception("No compose info file ...")

        print(nxc_build_file)
        oar_job = reserve_nodes(nb_nodes, site, cluster, walltime=15*60)
        self.oar_job_id, site = oar_job[0]
        roles_quantities = {"foo": ["foo", "bar"]}
        self.nodes = get_oar_job_nodes_nxc(self.oar_job_id, site, compose_info_file=nxc_build_file, roles_quantities=roles_quantities)
        print(self.nodes)

    def run(self):
        my_command = "echo \"Hello from $(whoami) at $(hostname) ($(ip -4 addr | grep \"/20\" | awk '{print $2;}'))\" > /tmp/hello"
        hello_remote = Remote(my_command, self.nodes["foo"], connection_params={'user': 'root'})
        hello_remote.run()

        my_command2 = "cat /tmp/hello"
        cat_remote = Remote(my_command2, self.nodes["foo"], connection_params={'user': 'root'})
        cat_remote.run()
        for process in cat_remote.processes:
            print(process.stdout)

def reserve_nodes(nb_nodes, site, cluster, walltime=3600):
    jobs = oarsub([(OarSubmission("{{cluster='{}'}}/nodes={}".format(cluster, nb_nodes), walltime, job_type=["allow_classic_ssh", "day"]), site)])
    return jobs

if __name__ == "__main__":
    ENGINE = MyEngine()
    try:
        ENGINE.start()
    except Exception as ex:
        print(f"Failing with error {ex}")
        oardel([(ENGINE.oar_job_id, None)])
        print("Giving back the resources")
